/* pdom.c */

/*
 *
 * Tarjan's fast dominator algorithm
 *
 * sophisticated version
 */

#include <stdio.h>
#include <assert.h>
#include <memory.h>
#include "cfg.h"
#include "set.h"
#include "tree.h"

#define PDOM(a,b) ((a)==0 ||(pdom1[(a)]<=pdom1[(b)] && pdom1[(b)]<=pdom2[(a)]))
#define SPDOM(a,b) (((a) != (b)) && PDOM((a),(b)))

static Node *node;
static int _PdomHt;

static int vertex[NMAX];
int ancestor[NMAX], child[NMAX], parent[NMAX];
int label[NMAX], semi[NMAX], size[NMAX];
INT_SET bucket[NMAX];
int n, pn;

void dfs(int v)
{
  int w;
  Edge *e;
  
  semi[v] = ++n;
  vertex[n] = label[v] = v;
  ancestor[v] = child[v] = 0;
  size[v] = 1;
  for (e = nPREDS (node[v]); e; e = eNEXTPRED (e))
    {
      w = eSRC (e);
      if (semi[w]==0)
	{
	  parent[w] = v;
	  dfs(w);
	}
    }
}

void compress(int v)
{
  if (ancestor[ancestor[v]] > 0)
    {
      compress(ancestor[v]);
      if (semi[label[ancestor[v]]] < semi[label[v]])
	label[v] = label[ancestor[v]];
      ancestor[v] = ancestor[ancestor[v]];
    }
}

int eval(int v)
{
  if (ancestor[v] == 0)
    return(label[v]);
  else
    {
      compress(v);
      if (semi[label[ancestor[v]]] >= semi[label[v]])
	return(label[v]);
      else
	return(label[ancestor[v]]);
    }
}

void link(int v, int w)
{
  int s, t;
  
  s = w;
  while (semi[label[w]] < semi[label[child[s]]])
    if (size[s] + size[child[child[s]]] >= 2*size[child[s]])
      {
	ancestor[child[s]] = s;
	child[s] = child[child[s]];
      }
    else
      {
	size[child[s]] = size[s];
	s = ancestor[s] = child[s];
      }
  label[s] = label[w];
  size[v] = size[v] + size[w];
  if (size[v] < 2*size[w])
    {
      t = child[v];
      child[v] = s;
      s = t;
    }
  while (s != 0)
    {
      ancestor[s] = v;
      s = child[s];
    }
}

int pdom_dfs(int v)
{
  int c, temp1;
  int temp = 0;

  pdom1[v] = ++n;
  for (c=pdom_child[v]; c; c=pdom_sibling[c])
    {parent[c] = v;
     temp1 = pdom_dfs(c);
     if (temp1 > temp) temp = temp1;
   }
  pdom2[v] = n;
  return (temp + 1);
}

int PdomHt ()
{return _PdomHt;
}

void postdom(Cfg * cfg)
{
  int i, u, v, w;
  Edge *ed;


  node = cfgNODES (cfg);

/*Zero out some pdom arrays. 
  memset (bucket,0, (ENDNum+1)*sizeof(INT_SET));
  memset (vertex,0, (ENDNum+1)*sizeof(int));
  memset (parent,0,(ENDNum+1)*sizeof(int)); 
  memset (pdom_sibling,0, (ENDNum+1)*sizeof(int)); */

  memset (semi,0, (ENDNum+1)*sizeof(int));
  memset (pdom,0, (ENDNum+1)*sizeof(int)); 


  n = pn = 0;
  dfs(ENDNum);
  
  size[0] = label[0] = semi[0] = 0;
  
  for (i=NumNodes; i>=2; i--)
    {
      w = vertex[i];
      
      for (ed = nSUCCS (node[w]); ed; ed = eNEXTSUCC (ed))
	{
	  v = eDST (ed);
	  u = eval(v);
	  if (semi[u] < semi[w])
	    semi[w] = semi[u];
	}
      bucket[vertex[semi[w]]] = add(w, bucket[vertex[semi[w]]]);
      link(parent[w], w);
      while (bucket[parent[w]] != NULLSET)
	{
	  v = VAL(bucket[parent[w]]);
	  u = eval(v);
	  pdom[v] = (semi[u] < semi[v] ? u : parent[w]);
	  bucket[parent[w]] = delete (bucket[parent[w]]);
	}
    }
  for (i=2; i<=NumNodes; i++)
    {
      w = vertex[i];
      if (pdom[w] != vertex[semi[w]])
	pdom[w] = pdom[pdom[w]];
    }
  pdom[ENDNum] = 0;

  /* create pdom_children lists */
  for (i=1; i<=ENDNum; i++)
    pdom_child[i] = 0;
  for (v=1; v<=ENDNum; v++)
    {
      pdom_sibling[v] = pdom_child[pdom[v]];
      pdom_child[pdom[v]] = v;
    }


  /*perform a dfs numbering of the pdom tree, and set parent array*/

  n = 0;  
  parent[ENDNum] = ENDNum+1;
  _PdomHt = 0;
  _PdomHt = pdom_dfs(ENDNum);

}

/* end of pdom.c */
